/* utils.c
 *
 *  Utility functions for xpsk31
 */

/*
 *  xpsk31: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card
 *
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "utils.h"

/*------------------------------------------------------------------------*/

/*  Save_QSO_Record()
 *
 *  Saves QSO record to .adif and .txt files
 */

  gboolean
Save_QSO_Record( void )
{
  /* For opening files */
  char file_name[64];

  /* Buffer for reading and printing to files */
  char buff[REC_BUFF_SIZE];
  char *buff_idx;

  int
	num_records, /* Number of QSO records    */
	num_logged,  /* Number of QSL's printed  */
	fopen_ret;   /* Open_File() return value */

  /* Create a file for QSO ADIF log if not open */
  Strlcpy( file_name, getenv("HOME"), sizeof(file_name) );
  Strlcat( file_name, "/xpsk31/log.adif", sizeof(file_name) );
  fopen_ret = Open_File( &adif_log_fp, file_name, "a" );
  if( fopen_ret == FILE_NEW )
  {
	snprintf( buff, sizeof(buff),
		"QSO Log file created in ADIF v1.0 format by %s <EOH>\n\n",
		PACKAGE_STRING );
	if( !File_Print(&adif_log_fp, buff) )
	  return( FALSE );
  }
  else if( fopen_ret == -1 )
	return( FALSE );

  /* Print QSO record to adif format file */
  snprintf( buff,  sizeof(buff),
	  "<CALL:%d>%s<QSO_DATE:8>%s<TIME_ON:4>%s\n"
	  "<FREQ:%d>%s<MODE:5>PSK31<RST_SENT:3>%3s<EOR>\n\n",
	  (int)strlen(qso_record.dx_call), qso_record.dx_call,
	  qso_record.date_adif, qso_record.time_adif,
	  (int)strlen(qso_record.freq), qso_record.freq,
	  qso_record.dx_rst );
  if( !File_Print(&adif_log_fp, buff) )
	return( FALSE );

  /* Create a file for station log if not already open */
  Strlcpy( file_name, getenv("HOME"), sizeof(file_name) );
  Strlcat( file_name, "/xpsk31/log.txt", sizeof(file_name) );
  fopen_ret = Open_File( &station_log_fp, file_name, "r+" );
  if( fopen_ret == FILE_NEW )
  {
	/* Print a header to log file */
	snprintf( buff,  sizeof(buff),
		_("|------------------------------------"\
		"--------------------------------------|\n"\
		"|         Station Log File - Created in"\
		" Text format by %10s          |\n"\
		"|------------------------------------"\
		"--------------------------------------|\n"),
		PACKAGE_STRING );
	if( !File_Print(&station_log_fp, buff) )
	  return( FALSE );

	/* Print an initial (0) count of records and qsl's */
	snprintf( buff,  sizeof(buff),
		_("|  Num of QSO Records:%-6d  Records Logged:%-6d"\
		"  Size of Records:%-4d  |\n"\
		"|------------------------------------"\
		"--------------------------------------|\n"),
		0, 0, RECORD_SIZE );
	if( !File_Print(&station_log_fp, buff) )
	  return( FALSE );
  }
  else if( fopen_ret == -1 )
	return( FALSE );

  /*** Increment and print Record and QSL count ***/
  /* Go to beginning of log file */
  fseek( station_log_fp, 0, SEEK_SET );

  /* Find line with "Number of QSO Records:" */
  if( !Locate_Line(&buff_idx, buff, _("QSO Records:"), station_log_fp) )
	return( FALSE );
  num_records = atoi(buff_idx);

  /* Find "Number of QSO's QSL'd:" string */
  if( !Locate_String(&buff_idx, buff, _("Records Logged:")) )
	return( FALSE );
  num_logged = atoi(buff_idx);

  /* Go back to beginning of line and print new record */
  if( fseek(station_log_fp, - (long)(1 + strlen(buff)), SEEK_CUR) != 0 )
	return( FALSE );
  snprintf( buff,  sizeof(buff),
	  _("|  Num of QSO Records:%-6d  Records Logged:%-6d"\
	  "  Size of Records:%-4d  |\n"\
	  "|------------------------------------"\
	  "--------------------------------------|\n"),
	  ++num_records, num_logged, RECORD_SIZE );
  if( !File_Print(&station_log_fp, buff) )
	return( FALSE );

  /*** Print record in text log file ***/
  if( fseek(station_log_fp, 0, SEEK_END) != 0 )
	return( FALSE );
  snprintf( buff,  sizeof(buff),
	  _("|DX/ CALL: %-14s NAME: %-12s QTH: %-12s  LOC: %-6s|\n"\
	  "|MY/ CALL: %-14s ZONE: %-12s QTH: %-12s  LOC: %-6s|\n"\
	  "|QSO DATE: %-11s    TIME: %-5s UTC   FREQ: %-13s QSL: NO    |\n"\
	  "|QSO MODE: %-11s DX-REPT: %-3s      MY-REPT: %-3s     VIA: %-12s|\n"\
	  "|MY TRANS: %-11s   POWER: %-9s    ANT: %-15s          |\n"\
	  "|MY RECVR: %-11s   N.FIG: %-7s      ANT: %-15s          |\n"\
	  "| REMARKS: %-45s                   |\n"\
	  "|------------------------------------"\
	  "--------------------------------------|\n"),
	  qso_record.dx_call, qso_record.dx_name, qso_record.dx_qth, qso_record.dx_loc,
	  rc_data.call, rc_data.zone, rc_data.qth, rc_data.loc,
	  qso_record.date, qso_record.time, qso_record.freq,
	  qso_record.mode, qso_record.dx_rst, qso_record.my_rst, qso_record.via,
	  rc_data.tx, rc_data.tx_power, rc_data.tx_ant,
	  rc_data.rx, rc_data.rx_nfig, rc_data.rx_ant, qso_record.remarks );
  if( !File_Print(&station_log_fp, buff) )
	return( FALSE );

  fflush( station_log_fp );
  fflush( adif_log_fp );

  ClearFlag( SAVE_RECORD );

  return( TRUE );
} /* Save_QSO_Record() */

/*------------------------------------------------------------------------*/

/*  Process_RST()
 *
 *  Processes RST-out and RST-in fields
 */

  void
Process_RST( GtkEditable *editable )
{
  G_CONST_RETURN gchar *rst;
  int len, idx;

  /* Validate RST: numerics only, */
  /* min 337 max 559, ignore blank field */
  rst = gtk_entry_get_text( GTK_ENTRY(editable) );
  len = (int)strlen(rst);

  /* Ignore blank fields */
  if( len == 0 ) return;

  for( idx = 0; idx < len; idx++ )
  {
	/* Reject non-numbers */
	if( (rst[idx] < '0') ||
		(rst[idx] > '9') )
	{
	  Bad_Entry_Dialog(
		  _("Invalid character entered\n"\
			"Numbers only allowed") );
	  return;
	}

	/* Reject RST < 337 or > 599 */
	switch( idx )
	{
	  case 0:
		if( (rst[idx] < '3') ||
			(rst[idx] > '5') )
		{
		  Bad_Entry_Dialog(
			  _("Number out of range\n"\
				"Min = 3 and Max = 5") );
		  return;
		}
		break;

	  case 1:
		if( rst[idx] < '3' )
		{
		  Bad_Entry_Dialog(
			  _("Number out of range\n"\
				"Min = 3 and Max = 9") );
		  return;
		}
		break;

	  case 2:
		if( rst[idx] < '7' )
		{
		  Bad_Entry_Dialog(
			  _("Number out of range\n"\
				"Min = 7 and Max = 9") );
		  return;
		}
	} /* switch( idx ) */

  } /* for( idx = 0; idx < len; idx++ ) */

} /* Process_RST() */

/*------------------------------------------------------------------------*/

/*  Select_Macro()
 *
 *  Selects a macro to be transmitted
 */

  void
Select_Macro( int macro_num )
{
  /* Transmit a macro */
  if( isFlagClear(TRANSMIT_MACRO) )
  {
	SetFlag( TRANSMIT_MACRO );
	ClearFlag( RECEIVE_MODE | TRANSMIT_KEYBD );
	gbl_macro_idx = macro_num - 1;
	Set_TxRx_Labels();

	/* Enable macro xmission */
	gtk_idle_add( Transmit_Macro, "dummy" );
  }
  else ClearFlag( TRANSMIT_MACRO );

} /* Select_Macro() */

/*------------------------------------------------------------------------*/

/*  Open_File()
 *
 *  Opens a file
 */

  int
Open_File( FILE **fp, char *fpath, const char *mode )
{
  struct stat buf;

  /* Check if file exixts */
  if( *fp == NULL )
  {
	/* If file does not exist, open new */
	if( stat(fpath, &buf) == -1 )
	{
	  char mesg[MESG_SIZE];

	  /* Create new file */
	  if( (*fp = fopen(fpath, "w+")) == NULL )
	  {
		perror( fpath );
		snprintf( mesg, sizeof(mesg),
			_("Failed to create\n%s\n"\
			  "Quit xpsk31 and correct"), fpath );
		Error_Dialog( mesg, QUIT );
		return( -1 );
	  }
	  else
	  {
		/* Re-open in requested mode */
		Close_File( fp );
		if( (*fp = fopen(fpath, mode)) == NULL )
		{
		  perror( fpath );
		  snprintf( mesg, sizeof(mesg),
			  _("Failed to open\n%s\n"\
				"Quit xpsk31 and correct"), fpath );
		  Error_Dialog( mesg, QUIT );
		  return( -1 );
		}
		else return( FILE_NEW );
	  }
	} /* if( stat(fpath, buf) == -1 ) */
	else /* If file exists, re-open in required mode */
	{
	  if( (*fp = fopen(fpath, mode)) == NULL )
	  {
		char mesg[MESG_SIZE];
		perror( fpath );
		snprintf( mesg, sizeof(mesg),
			_("Failed to open\n%s\n"\
			  "Quit xpsk31 and correct"), fpath );
		Error_Dialog( mesg, QUIT );
		return( -1 );
	  }
	  else return( FILE_OPEN );

	} /* else */

  } /* if( *fp == NULL ) */

  return( FILE_OPEN );
} /* Open_File() */

/*------------------------------------------------------------------------*/

/*  Close_File()
 *
 *  Closes and NULL's an open file pointer
 */

  void
Close_File( FILE **fp )
{
  if( *fp != NULL )
  {
	fclose( *fp );
	*fp = NULL;
  }
}

/*------------------------------------------------------------------------*/

/*  File_Print()
 *
 *  Prints a string to a file
 */

  gboolean
File_Print( FILE **fp, char *string )
{
  if( fprintf(*fp, "%s", string) != (int)strlen(string) )
  {
	perror( "xpsk31: fprintf()" );
	ClearFlag( SAVE_RECORD );
	Error_Dialog(
		_("Error printing to file\n"\
		  "Quit xpsk31 and correct"), QUIT );
	Close_File( fp );
	return( FALSE );
  }

  /* Flush file */
  if( fflush(*fp) != 0 )
  {
	perror( "xpsk31: fflush()" );
	ClearFlag( SAVE_RECORD );
	Error_Dialog(
		_("Failed to fflush file\n"\
		  "Quit xpsk31 and correct"), QUIT );
	return( FALSE );
  }

  return( TRUE );
} /* File_Print() */

/*------------------------------------------------------------------------*/

/*  Locate_Line()
 *
 *  Finds a line in a text file containing a given string
 */

  gboolean
Locate_Line( char **line_idx, char *line, const char *string, FILE *fp )
{
  int err;

  /* Load lines and look for string */
  do
  {
	err = Load_Line( line, fp, _("line in Log File") );
	*line_idx = strstr(line, string);
  }
  while( (err == 0) && (*line_idx == NULL) );

  /* Usually EOF is the cause of error */
  if( err )
  {
	Error_Dialog(
		_("Error searching text file\n"\
		  "May be empty or corrupted"), QUIT );
	return( FALSE );
  }

  *line_idx += strlen(string);

  return( TRUE );
} /* Locate_Line() */

/*------------------------------------------------------------------------*/

/*  Find_String()
 *
 *  Finds a string following spaces
 */

  gboolean
Locate_String( char **line_idx, char *line, const char *string )
{
  if( (*line_idx = strstr(line, string)) == NULL )
  {
	Error_Dialog(
		_("Error searching station_log.txt\n"\
		  "May be empty or corrupted"), QUIT );
	return( FALSE );
  }

  *line_idx += strlen(string);

  return( TRUE );
} /* Find_String() */

/*------------------------------------------------------------------------*/

/*  Read_QSO_Record()
 *
 *  Reads and validates QSO record entries
 */

  gboolean
Read_QSO_Record( void )
{
  GtkWidget *entry;

  /* Enter field values to QSO record structure */
  entry = lookup_widget( main_window, "callsign" );
  Strlcpy( qso_record.dx_call,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.dx_call) );
  entry = lookup_widget( main_window, "rst_in" );
  Strlcpy( qso_record.my_rst,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.my_rst) );
  entry = lookup_widget( main_window, "rst_out" );
  Strlcpy( qso_record.dx_rst,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.dx_rst) );
  entry = lookup_widget( main_window, "op_name" );
  Strlcpy( qso_record.dx_name,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.dx_name) );
  entry = lookup_widget( main_window, "qth_name" );
  Strlcpy( qso_record.dx_qth,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.dx_qth) );
  entry = lookup_widget( main_window, "locator" );
  Strlcpy( qso_record.dx_loc,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.dx_loc) );
  entry = lookup_widget( main_window, "band" );
  Strlcpy( qso_record.freq,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.freq) );

  /* Enter operating mode */
  if( isFlagSet(MODE_BPSK) )
	Strlcpy( qso_record.mode, "BPSK", sizeof(qso_record.mode) );
  else
	Strlcpy( qso_record.mode, "QPSK", sizeof(qso_record.mode) );

  /* Validate QSO Record */
  if( (strlen(qso_record.dx_call) > 2) &&
	  (strlen(qso_record.my_rst) == 3) &&
	  (strlen(qso_record.freq)    > 0) )
	return( TRUE );
  else return( FALSE );

} /* Read_QSO_Record() */

/*------------------------------------------------------------------------*/

/***  Memory allocation/freeing utils ***/
gboolean mem_alloc( void **ptr, size_t req )
{
  free_ptr( ptr );
  *ptr = malloc( req );
  if( *ptr == NULL )
  {
	perror( "xpsk31: alloc():" );
	Error_Dialog( _("A memory allocation failed\n"\
		  "Please quit xpsk31 and correct"), QUIT  );
	return( FALSE );
  }
  return( TRUE );
} /* End of void mem_alloc() */

/*------------------------------------------------------------------------*/

gboolean mem_realloc( void **ptr, size_t req )
{
  *ptr = realloc( *ptr, req );
  if( *ptr == NULL )
  {
	perror( "xpsk31: realloc():" );
	Error_Dialog( _("A memory re-allocation failed\n"\
		  "Please quit xpsk31 and correct"), QUIT  );
	return( FALSE );
  }
  return( TRUE );
} /* End of void mem_realloc() */

/*------------------------------------------------------------------------*/

void free_ptr( void **ptr )
{
  if( *ptr != NULL )
	free( *ptr );
  *ptr = NULL;

} /* End of void free_ptr() */

/*------------------------------------------------------------------------*/

/* Strlcpy()
 *
 * Copies n-1 chars from src string into dest string. Unlike other
 * such library fuctions, this makes sure that the dest string is
 * null terminated by copying only n-1 chars to leave room for the
 * terminating char. n would normally be the sizeof(dest) string but
 * copying will not go beyond the terminating null of src string
 */
  void
Strlcpy( char *dest, const char *src, size_t n )
{
  char ch = src[0];
  int idx = 0;

  /* Leave room for terminating null in dest */
  n--;
 
  /* Copy till terminating null of src or to n-1 */
  while( (ch != '\0') && (n > 0) )
  {
	dest[idx] = src[idx];
	idx++;
	ch = src[idx];
	n--;
  }

  /* Terminate dest string */
  dest[idx] = '\0';

} /* Strlcpy() */

/*------------------------------------------------------------------*/

/* Strlcat()
 *
 * Concatenates at most n-1 chars from src string into dest string.
 * Unlike other such library fuctions, this makes sure that the dest
 * string is null terminated by copying only n-1 chars to leave room
 * for the terminating char. n would normally be the sizeof(dest)
 * string but copying will not go beyond the terminating null of src
 */
  void
Strlcat( char *dest, const char *src, size_t n )
{
  char ch = dest[0];
  int idd = 0; /* dest index */
  int ids = 0; /* src  index */

  /* Find terminating null of dest */
  while( (n > 0) && (ch != '\0') )
  {
	idd++;
	ch = dest[idd];
	n--; /* Count remaining char's in dest */
  }

  /* Copy n-1 chars to leave room for terminating null */
  n--;
  ch = src[ids];
  while( (n > 0) && (ch != '\0') )
  {
	dest[idd] = src[ids];
	ids++;
	ch = src[ids];
	idd++;
	n--;
  }

  /* Terminate dest string */
  dest[idd] = '\0';

} /* Strlcat() */

/*------------------------------------------------------------------*/

